#
#
#   Blockwise: a tcl/tk block wise programming and interaction
#              environment, consisting of a scrollable garphical 
#              workspace containing graphical "block" representations 
#              of various types of processes or program items
#
#
#  This is a trial version (V0.1) to see what is a valid environment 
#  and how it can be utilized and give an improved interaction interface
#  and improved programming effort throughput.
#
#  Some of the concepts are examplified and tried out on the fly, so 
#  this is not a stabelized complete concept (yet), but sufficiently 
#  well known parts are available and tested to guarantee added value.
#  Not all I have tried is already incorporated here.
#


# start main windows

set mw {}
set mc $mw.c
set bcount 0
set scopeindex 0
set wireindex 0
set shellindex 0
set drumindex 0


proc main_window {root lproject} {
   global mw mc project
   set mw $root.mw
   set mc $mw.c
   set project $lproject
   
   catch {destroy $mw [winfo children $mw]}
   if {$root == ""} {
      wm title . Bwise
      wm geometry . 500x350
   }
   frame $mw -width 400 -height 300;
      pack $mw -side bottom  -expand y -fill both -anchor s
   scrollbar $mw.hscroll -orient horiz -command "$mc xview"
   scrollbar $mw.vscroll -command "$mc yview"
   canvas $mc -width 1000 -height 800 -bg gray80 \
      -scrollregion {0 0 1000 800} \
      -xscrollcommand "$mw.hscroll set" \
      -yscrollcommand "$mw.vscroll set"
   pack $mw.vscroll -side right -fill y
   pack $mw.hscroll -side bottom -fill x
   pack $mc -side right -expand n -fill both
}

proc newblock { {name {}} \
                {x 0} {y 0} {w 50} {h 50} \
                {pinsin {in1 in2}} {pinsout {}}\
                {moretags ""}
              } \
{
   global bcount mw mc
   if {$name == ""} {set name "block$bcount" ; set bcount [expr $bcount +1]}
   eval $mc create rect $x $y [expr $x+$w] [expr $y+$h] \
      -tags "{$name newblock block $moretags}" -fill yellow -outline darkblue
   eval $mc create text  [expr $x+$w/2] [expr $y+$h] \
      -anchor n -text $name -fill darkblue -tags "{$name crb label $moretags}"
   set ii 0
   foreach i $pinsin {
      eval $mc create text [expr $x-1] [expr $y+19+$ii] \
         -anchor se -text $i -tags "{$name crb pinname $i $moretags}"
      eval $mc create line \
         [expr $x-20] [expr $y+20+$ii] [expr $x+0] [expr $y+20+$ii] \
         -width 2 -fill darkblue  -tags "{$name newblock pin $i typein $moretags}"
      set ii [expr $ii+15]
   }
   set ii 0
   foreach i $pinsout {
      eval $mc create text [expr $x+$w+1] [expr $y+19+$ii] \
         -anchor sw -text $i -tags "{$name crb pinname $i $moretags}"
      eval $mc create line \
         [expr $x+$w+20] [expr $y+20+$ii] [expr $x+$w+0] [expr $y+20+$ii] \
         -width 2 -fill darkblue  -tags "{$name newblock pin $i typeout $moretags}"
      set ii [expr $ii+15]
   }
}


proc  createshell {{name shell}} {
   global mc shellindex
   
   if {$name == "shell"} {set name $name$shellindex;
      set shellindex [expr $shellindex+1]
   }
   newblock $name 0 0 200 120 {command ouput do}
   frame $mc.$name ;
   eval $mc create window 20 10 -anchor nw -height 100 -width 170 \
      -tags "{$name crb shellwindow}" -window $mc.$name

   set t $name; append t _entry
   entry $mc.$name.e -textvar $t
   puts 1
   eval set ttt   "{$mc.$name.t insert end \"\\n\$$t\\n\" tred;  $mc.$name.t insert end  \"\[eval \$$t\]\" tnavy; $mc.$name.t see end; $mc.$name.e selection range 0 end}"
#   puts ,$ttt,
   button $mc.$name.b -text Eval -command \
      "$ttt" \
      -borderw 0 -padx 0 -pady 0
#      "$mc.$name.t insert end \n\\\$$t\\n tred;  $mc.$name.t insert end  [eval \\\$$t]; $mc.$name.t see end; $mc.$name.e select 0 end" \
# '{eval \$name"_entry}"
   pack $mc.$name.e -side bottom -anchor s -expand n -fill x
   pack $mc.$name.b -side bottom -anchor s -expand n -fill x
   text $mc.$name.t
   pack $mc.$name.t -side top -anchor n -expand n -fill both
   $mc.$name.t tag configure  tred  -foreground red -font "helvetica 8"
   $mc.$name.t tag configure  tnavy  -foreground navy -font "helvetica 8"
   bind $mc.$name.e <Return> "eval  $mc.$name.b invoke"
   
}


proc createscope {{name scope}} {
   global mc scopeindex
   
   if {$name == "scope"} {set name $name$scopeindex;
      set scopeindex [expr $scopeindex+1]
   }
   newblock $name 0 0 200 120 {in trigger}
   frame $mc.$name
   eval $mc create window 20 10 -anchor nw -height 100 -width 170 \
      -tags "{$name crb scopewindow}" -window $mc.$name

   canvas $mc.$name.c -height 10 -width 10 -bg gray10
   pack $mc.$name.c -expand y -fill both
   $mc.$name.c create line 0 40 170 40 -fill gray30
   for {set i 1} {$i<1000} {set i [expr $i+1]} { \
      eval $mc.$name.c create line [expr $i-1] \
         [expr 40-(40/($i/100.0))*sin(2*3.1415*($i-1)/30 )] $i \
         [expr 40-(40/($i/100.0))*sin(2*3.1415*$i/30 ) ] -fill lightblue -tags "{$name createscope channel1}"
   }
   eval scrollbar  $mc.$name.s -orient horiz -command "{$mc.$name.c xview}"
   eval $mc.$name.c conf -xscrollcommand "{$mc.$name.s set}" -scrollregion  "{0 0 1000 100}"
   pack $mc.$name.s -side bottom -expand 0 -fill x
}


proc gen_drumtrack { {bl {drum}} {tr {bd hho hhc sd ride crash tamb}} {steps 16} }   {
   global mc drumindex
   
   if {$bl == "drum"} {set bl $bl$drumindex;
      set drumindex [expr $drumindex+1]
   }
   
#   newblock $bl 0 0 220 120 {trigger_in trigger_out command}
#   $mc create rect 10 10 210 110 -fill white -tags "$bl newblock workp"
   set ym [expr [llength $tr]-1]
   newblock $bl 0 0 [expr 36+8*($steps)] [expr 60+$ym*20 ] {trigger_in trigger_out command}
   $mc create rect 10 10 [expr 36+8*($steps)-10] [expr 60+$ym*20-10 ] -fill white -tags "$bl newblock workp"

   set j 0
      for {set i 0} {$i < $steps} {set i [expr $i+4]} {
          $mc create line [expr 21 + 8*$i] 10 \
             [expr 21 + 8*$i]  [expr 60+$ym*20-10 ] -fill blue \
             -tags "$bl sepa"
      }
   foreach n $tr {
      $mc create line 10 [expr $j*20+30] [expr 36+8*($steps)-10] [expr $j*20+30] -fill red -tags "$bl workp $n"
      $mc create text 10 [expr 20*$j+27] -text "$n" -tags "$bl workp  text$n"  -fill blue -anchor sw
      for {set i 0} {$i < $steps} {set i [expr $i+1]} { \
         $mc create oval [expr 18 + 8*$i] [expr 20*$j+27] \
             [expr 24 +8*$i] [expr 20*$j+33] -fill red -outline {} \
             -tags "$bl drumbol drtr$n drtr$i" ; \
      }
      set j [expr $j+1]
   }
}

proc drumscan { {bl {drum0}} {tr {bd hho hhc sd}} } {
   global mc
   set maxi 0
   foreach k [listunion [$mc find withtag drumbol] [$mc find withtag $bl] ] \
   {
      if {[$mc itemcget $k -outline] != {}} {
          eval lappend outlist \{ \
           [string range [lindex [$mc itemcget $k -tag] 2]  4 end] \
           [string range [lindex [$mc itemcget $k -tag] 3]  4 end] \}
      }
      set ii [string range [lindex [$mc itemcget $k -tag] 3]  4 end]
      if {$ii > $maxi} {set maxi $ii}
   }
   set outlist [lsort -index 1 -int $outlist]
   
   set argi "len [expr 1000+(1000/8)*$maxi] 0 0\n"
   foreach i $outlist {
      append argi "s_[lindex $i 0] [expr [lindex $i end] * 2000.0 / 16]"
      switch [lindex $i 0] {
         bd    {append argi " 200 0.9"}
         sd    {append argi " 300 0.7"}
         hhc   {append argi " 100 0.3"}
         hho   {append argi " 200 0.15"}
         ride  {append argi " 300 0.8"}
         crash {append argi " 300 0.8"}
         tamb  {append argi " 100 0.3"}
          default   {append argi " 200 0.3"}
      }
      append argi "\n"
   }
#   set f [open $bl.txt w]
#      puts $f "len 2000"
#      foreach i $outlist {
#         puts $f "s_[lindex $i 0] [expr [lindex $i end] * 2000.0 / 16]"
#      }
#   close $f
   return $argi
}

proc delete_selblocks {} {
   global mc;
   foreach i [listunion [$mc find withtag selection0] [$mc find withtag bb] ] { \
       foreach j [lindex [$mc itemcget $i -tags] 0] {
          block_del $j ;
          foreach k [$mc find withtag wire] {
             if {[lsearch [$mc itemcget $k -tag] $j] > 0} {
                $mc del $k
             }
          }
       }
    }
}

proc block_del  n {
   global mc;
   foreach i [$mc find withtag $n] {
      if {[lindex [ $mc itemcget $i -tag ] 0] == $n}  {$mc del $i}
   }
}

proc genwire {} {
   global mc
   set t [lrange [$mc find withtag selectedpin] 0 1]
   if {[llength $t] == 2} { 
         connect {}    [lindex [$mc itemcget [lindex $t 0] -tags] 0] \
                       [lindex [$mc itemcget [lindex $t 0] -tags] 3] \
                       [lindex [$mc itemcget [lindex $t 1] -tags] 0] \
                       [lindex [$mc itemcget [lindex $t 1] -tags] 3]
   }
}

proc mainbuttons {root} {
   global mw
   set mb $root.fb
   catch {destroy $mb}
   frame $mb 
      pack $mb -side top -fill x -expand y -anchor n
   button $mb.bnewb -text "New Block" -command {newblock {} }
   pack $mb.bnewb -side left -fill none -expand n
   if {$root == ""} \
         {global tcl_interative; button $mb.quit -text "Quit" -command {eval destroy [winfo children .]; \
            if {$tcl_interactive == 0} {destroy .}  }  } \
         {button $mb.quit -text "Quit" -command "eval destroy [winfo children $root]" }
   pack $mb.quit -side right -fill none -expand n
   
   image create photo paperim -file paper.gif
   button .fb.p -image paperim -borderw 0 -padx 0 -pady 0 -relief flat -command createshell
   pack .fb.p -side left
   button $mb.bwire -text "Wire" -command genwire -fg green
      pack $mb.bwire -side left -fill none -expand n
   
   button $mb.bcdrum -text "Drum" -command gen_drumtrack
   button $mb.bcscope -text "Scope" -command createscope
   pack $mb.bcdrum -side left
   pack $mb.bcscope -side left
   button $mb.bscan -text "Scan Drum" -command \
      {exec drumip.exe drum0.wav << [drumscan] &}
   pack $mb.bscan -side right -fill none -expand 0
   button $mb.bdel -text "Del Sel" -command delete_selblocks
   pack $mb.bdel -side right -expand n -fill none
   button $mb.bsave -text "Save" -command save_canvas
   pack $mb.bsave -side right -fill none -expand 0
   button $mb.bload -text "Load" -command {reload_canvas}
   pack $mb.bload -side right -fill none -expand 0
}

proc wire_coord {name pin x y} {
   global mc
#   puts $name,$pin,$x,$y;
   set w  [ listunion [$mc find withtag $pin] [listunion [$mc find withtag $name] \
       [$mc find withtag wire]] ] ;
#    foreach i [ listunion [$mc find withtag wire] [$mc find withtag $name] ] {
#
#    }
    if {$w == {}} { return}
    foreach ww $w {
     set t [eval $mc itemcget $ww -tags]
#     puts $ww,$t
     for {set i 3} {$i < [llength $t]} {set i [expr $i+2]} {
#       puts $i
       if {[lindex $t $i] == $name} {
          if {[lindex $t [expr $i+1]] == $pin} {
             eval $mc coords $ww \
                "[ lreplace [$mc coords $ww] [expr $i-3] [expr $i+1-3] $x $y]" ; 
          } 
       } ; 
    } ;
    }
}
#         i [lrange [eval $mc itemcget $w -tags] 3 end] {
#    set c [expr { [lsearch -exact [eval $mc itemcget $w -tags] $pin]-4} ];
#    eval $mc coords $w [ lreplace [eval $mc coords $w] $c [expr $c+1] $x $y] ; 

proc find2l {l1 l2} {
   global mc
   foreach i [$mc find withtag $l1] { \
      if {[lsearch [$mc itemcget $i -tags]  $l2 ] >= 0} \
         {return $i}
   } 
}


proc listunion {a b} {
   global mc
   set l {}; 
   foreach i $a  { \
      if {[lsearch -exact $b $i] >= 0} \
         {lappend l $i} \
   } ;
   return $l
}


proc wire_update name { 
   global mc
   foreach i [ listunion [$mc find withtag $name] [$mc find withtag pin ] ] \
      {eval  wire_coord $name [lindex [eval $mc itemcget $i -tags] 3] \
      [lrange [eval  $mc coords $i] 0 1] \
   }
}

# find all items that have all tags from the list tl

proc tag_and tl {
   global mc
   set r [$mc find withtag [lindex $tl 0]]
   foreach i [lrange $tl 1 end] {
      set r [listunion $r [$mc find withtag $i] ]
   }
   return $r
}


proc connect {wire name1 pin1 name2 pin2} { 
   global mc wireindex
#   puts $wire,$name1,$pin1
   set i1 [eval tag_and "{$name1 $pin1 pin}"]
   set i2 [eval tag_and "{$name2 $pin2 pin}"]
#   set i1 [find2l $name1 $pin1];
#   set i2 [find2l $name2 $pin2];
#   puts $i1,$i2;  
   set x1 [lindex [$mc coords $i1] 0];
   set y1 [lindex [$mc coords $i1] 1];
   set x2 [lindex [$mc coords $i2] 0];
   set y2 [lindex [$mc coords $i2] 1];
#   puts $x1,$y1,$x2,$y2;
   set match {}
   set t1 "{[lsort [list $name1 $pin1 $name2 $pin2]]}"
   foreach i [$mc find withtag wire] {
      set t2 "{[lsort [lrange [$mc itemcget $i -tags] 3 6]]}"
      if { $t1 == $t2 } {
         lappend match [lindex [$mc itemcget $i -tags] 0]
      }
   }
#   puts $match
   if {$match == {}} {
      if {$wire == ""} {
         set wire wire$wireindex
         set wireindex [expr $wireindex+1]
      }
      eval $mc create line $x1 $y1 $x2 $y2 -fill darkblue \
         -tags "{$wire connect wire $name1 $pin1 $name2 $pin2}"
      set i [eval tag_and "{$name1 $pin1 pin}"]
      $mc itemco $i -fill darkblue
      $mc dtag $i selectedpin
      set i [eval tag_and "{$name2 $pin2 pin}"]
      $mc itemco $i -fill darkblue
      $mc dtag $i selectedpin
   } {
         foreach i $match { $mc del $i}
   }
}

proc gen_netlist {} {
   global mc
   set o ""
   foreach w [tag_and wire] {
      append o "connect [lindex [$mc itemcget $w -tags] 0] [lrange [$mc itemcget $w -tags] 3 6] \n"
   }
   return $o
}

proc cbbox {{name all} {selection selection0}} {
   global mc;
   set tl {};
   foreach i [$mc find withtag $name] { \
      if {[lsearch -exact [$mc itemcget $i -tags] wire ] < 0} \
         {lappend tl $i} 
   };
   set t [eval $mc create rect [eval $mc bbox $tl] \
      -fill red -tags "{$name bbox $selection}" ];
   $mc lower $t;set lx [$mc coords $t];
   eval $mc create text [expr ([ lindex $lx 0 ]+[ lindex $lx 2 ])/2 ] \
      [expr [lindex $lx 3]+1] -text $name \
      -tags "{$name bb blockname bbox $selection}"  \
      -anchor n -fill red -font "{helvetica 12 bold}"
}

proc togglepinsel ids {
   global mc
   foreach id $ids {
      if {[$mc itemcget $id -fill] != "green"} {
         $mc itemco $id -fill green
         $mc addtag selectedpin withtag $id
      } {
         $mc itemco $id -fill darkblue
         $mc dtag $id selectedpin
      }
   }
}

proc blockbinds {} {
   global mc
   $mc bind all <ButtonPress-1> { \
      global hx hy hl mc; set hx %X; set hy %Y; \
      set hl [lindex [$mc gettags current] 0]; \
      bind $mc <Motion> { \
         global hx hy hl mc; \
         foreach i [$mc find withtag "$hl"] { \
            if {[lindex [$mc itemcget $i -tags] 0 ] == "$hl"} \
               {$mc move  $i  [expr %%X-$hx] [expr %%Y-$hy]} \
         }; \
         wire_update $hl ; set hx %%X; set hy %%Y; \
      } \
   } ;
   bind $mc <ButtonRelease-1> {global mc; bind $mc <Motion> {}} ;
   $mc bind block <Double-ButtonPress-1> {  \
      if { [llength \
            [listunion \
               [$mc find withtag \
                  [lindex [$mc itemcget current -tags] 0] \
               ] [ \
                  $mc find withtag bbox \
               ] \
            ] \
           ] == 0 \
      } \
      { cbbox [lindex [$mc itemcget current -tags] 0] selection0} \
      { eval $mc del  [listunion [$mc find withtag bbox] \
         [$mc find withtag [lindex [$mc itemcget current -tags] 0] ]]} ; \
   }
   $mc bind pin <ButtonPress-1> {
      togglepinsel current
   }
   $mc bind wire <ButtonPress-1> {
      togglepinsel [tag_and [ concat [lrange  \
         [$mc itemcget current -tags] 3 4]  pin]];
      togglepinsel [tag_and [ concat [lrange  \
         [$mc itemcget current -tags] 5 6]  pin]] }
   $mc bind  drumbol <ButtonPress-1> {
      global mc;
      if {[$mc itemcget current -outline] == {}} \
         {$mc itemco current -outline purple -width 2}  \
         {$mc itemco current -outline {}}
      }

}
#      set l2 [$mc find withtag current]; \
#      .we.f2.l1 conf -text [ lrange [$mc itemcget $l2 -tags] 0 0 ] ; \
#      set e2 [$mc itemcget current -fill]; \
#      bind .we.f2.e1 <Return> { $mc itemco $l2 -fill $e2 } ; \
#      .we.f2.e1 selection range 0 end \
#   }

proc wire_other {wname block pin} {
   global mc; set r {}; 
   set t [$mc itemcget [eval "tag_and {$wname wire}"] -tags]; 
   if {[lindex $t 3] == $block && [lindex $t 4] == $pin} \
       {set r [lrange $t 5 6]};
   if {[lindex $t 5] == $block && [lindex $t 6] == $pin} \
      {set r [lrange $t 3 4]};
   return $r
}

proc pin_get_wirenames {block pin} {
   global mc;
   set o {}; 
   foreach i [eval "tag_and {wire $block $pin}"] {
      eval "set w {[$mc itemcget $i -tags]}";
      if {[lrange $w 3 4] == [list $block $pin] || [lrange $w 5 6] ==  [list $block $pin]} {
         lappend o [lindex $w 0]}
      }; 
   return $o
}

proc block_get_pinnames {block {pintype {}} } {
   global mc;
   set o {};
   foreach i [eval "tag_and {$block pin $pintype}"] {
      lappend o [lindex [$mc itemcget $i -tags] 3]
   };
   return $o
}

proc showactive {block pin c} {
   global mc;
   if {$c == {}} {return};
   $mc itemco [ eval "tag_and {pin $block $pin}"] -fill $c
}

proc flood {block pin} {
   global mc;
   set delay 1000;
   if {$block == {} && $pin == {} } {return {}};
   showactive $block $pin red; update;
   foreach w [pin_get_wirenames $block $pin] {
      set nb [lindex [wire_other $w $block $pin] 0];
      foreach o [block_get_pinnames $nb typeout] {
         eval "after $delay {flood $nb $o}"
      }
   };
   eval "after $delay showactive $block $pin navy"; update
}

proc block_name_fromid id {
   global mc
   eval return [lindex [$mc itemcget $id -tags] 0]
}

proc netlistout {{tag {}}} {
   set o {};
   foreach id [eval tag_and "{block $tag}"] {
      set bl [block_name_fromid $id];
      foreach pin [block_get_pinnames $bl typeout] {
         foreach wire [pin_get_wirenames $bl $pin] {
            set ot [wire_other $wire $bl $pin];
            if {$ot != {}} {
               lappend o "$bl $pin $ot"
               # puts "$bl $pin $ot"
            }
         }
      }
   }
   return $o;
}

proc savehis {file} {
   set f [open $file w];
   puts $f [history];
   close $f
}


proc save_canvas {} {
   global mc
   set o ""
   set sv {bcount scopeindex wireindex shellindex drumindex}
   eval global $sv
   eval {append o global " " $sv \n}
   foreach i $sv {
      eval set j $$i
      append o set " " $i " " $j \n
   }
   foreach i [$mc find all] {
      if {[$mc type $i] != "window"} {
         append o $mc " " create " " [$mc type $i] " " [$mc coords $i]
         foreach k [ $mc itemco $i] {
            append o " " [lindex $k 0] " " [list [lindex $k end]]
         }
        append o \n\
     }
   }
   set fn [tk_getSaveFile];
   if {$fn != ""} {
      set fd [open $fn w];
      puts $fd $o; close $fd 
   }
}

proc reload_canvas {} {
   global mc
   set fn [tk_getOpenFile];
   if {$fn != ""} {
      $mc del all
      foreach i [winfo children $mc] {destroy $i}
      source $fn
   }
}



############################ MAIN CODE  #################################

mainbuttons ""
main_window {} Main
blockbinds

history keep 1000
